//
// Copyright (c) 2010 Thomas Klimpel
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_NUMERIC_BINDINGS_WRAPPER_HPP
#define BOOST_NUMERIC_BINDINGS_WRAPPER_HPP

#include <boost/numeric/bindings/column.hpp>
#include <boost/numeric/bindings/conj.hpp>
#include <boost/numeric/bindings/herm.hpp>
#include <boost/numeric/bindings/lower.hpp>
#include <boost/numeric/bindings/noop.hpp>
#include <boost/numeric/bindings/row.hpp>
#include <boost/numeric/bindings/symm.hpp>
#include <boost/numeric/bindings/trans.hpp>
#include <boost/numeric/bindings/unit_lower.hpp>
#include <boost/numeric/bindings/unit_upper.hpp>
#include <boost/numeric/bindings/upper.hpp>
#include <boost/numeric/bindings/vector_view.hpp>

#endif

