//
// Copyright (c) 2009 Rutger ter Borg
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_NUMERIC_BINDINGS_BLAS_LEVEL2_HPP
#define BOOST_NUMERIC_BINDINGS_BLAS_LEVEL2_HPP

#include <boost/numeric/bindings/blas/level2/gbmv.hpp>
#include <boost/numeric/bindings/blas/level2/gemv.hpp>
#include <boost/numeric/bindings/blas/level2/gerc.hpp>
#include <boost/numeric/bindings/blas/level2/ger.hpp>
#include <boost/numeric/bindings/blas/level2/geru.hpp>
#include <boost/numeric/bindings/blas/level2/hbmv.hpp>
#include <boost/numeric/bindings/blas/level2/hemv.hpp>
#include <boost/numeric/bindings/blas/level2/her2.hpp>
#include <boost/numeric/bindings/blas/level2/her.hpp>
#include <boost/numeric/bindings/blas/level2/hpmv.hpp>
#include <boost/numeric/bindings/blas/level2/hpr2.hpp>
#include <boost/numeric/bindings/blas/level2/hpr.hpp>
#include <boost/numeric/bindings/blas/level2/sbmv.hpp>
#include <boost/numeric/bindings/blas/level2/spmv.hpp>
#include <boost/numeric/bindings/blas/level2/spr2.hpp>
#include <boost/numeric/bindings/blas/level2/spr.hpp>
#include <boost/numeric/bindings/blas/level2/symv.hpp>
#include <boost/numeric/bindings/blas/level2/syr2.hpp>
#include <boost/numeric/bindings/blas/level2/syr.hpp>
#include <boost/numeric/bindings/blas/level2/tbmv.hpp>
#include <boost/numeric/bindings/blas/level2/tbsv.hpp>
#include <boost/numeric/bindings/blas/level2/tpmv.hpp>
#include <boost/numeric/bindings/blas/level2/tpsv.hpp>
#include <boost/numeric/bindings/blas/level2/trmv.hpp>
#include <boost/numeric/bindings/blas/level2/trsv.hpp>

#endif
