//
// Copyright (c) 2009 Rutger ter Borg
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_NUMERIC_BINDINGS_DETAIL_CONVERT_TO_HPP
#define BOOST_NUMERIC_BINDINGS_DETAIL_CONVERT_TO_HPP

namespace boost {
namespace numeric {
namespace bindings {
namespace detail {

template< typename Target, typename Source >
struct convert_to {};

} // namespace detail
} // namespace bindings
} // namespace numeric
} // namespace boost

#endif
