//
// Copyright (c) 2002--2010
// Toon Knapen, Karl Meerbergen, Kresimir Fresl,
// Thomas Klimpel and Rutger ter Borg
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// THIS FILE IS AUTOMATICALLY GENERATED
// PLEASE DO NOT EDIT!
//

#ifndef BOOST_NUMERIC_BINDINGS_LAPACK_AUXILIARY_HPP
#define BOOST_NUMERIC_BINDINGS_LAPACK_AUXILIARY_HPP

#include <boost/numeric/bindings/lapack/auxiliary/labrd.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lacgv.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lacon.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/laebz.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lalsd.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/langb.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lange.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lanhb.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lanhe.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lanhp.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lanhs.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lansb.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lansp.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lansy.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lantb.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lantp.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/lantr.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larf.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larfb.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larfg.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larft.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larfx.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/largv.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larnv.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larrb.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larre.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/larz.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/latrd.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/latrs.hpp>
#include <boost/numeric/bindings/lapack/auxiliary/latrz.hpp>

#endif
