
#ifndef _util_misc_math_h
#define _util_misc_math_h

#include <math.h>

#ifndef M_PI
#  define M_PI           3.14159265358979323846  /* pi */
#endif

#ifndef M_LN2
#  define M_LN2          0.69314718055994530942  /* log_e 2 */
#endif

#ifndef M_PI_2
#  define M_PI_2         1.57079632679489661923  /* pi/2 */
#endif

#endif
