//
// Copyright (c) 2009 Rutger ter Borg
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_NUMERIC_BINDINGS_STD_HPP
#define BOOST_NUMERIC_BINDINGS_STD_HPP

#include <boost/numeric/bindings/std/vector.hpp>
#include <boost/numeric/bindings/std/valarray.hpp>

#endif
