//
// predicate.h
//
// Copyright (C) 2009 Edward Valeev
//
// Author: Edward Valeev <evaleev@vt.edu>
// Maintainer: EV
//
// This file is part of the SC Toolkit.
//
// The SC Toolkit is free software; you can redistribute it and/or modify
// it under the terms of the GNU Library General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// The SC Toolkit is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public License
// along with the SC Toolkit; see the file COPYING.LIB.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The U.S. Government is granted a limited license as per AL 91-7.
//

#ifndef _mpqc_src_lib_math_scmat_predicate_h
#define _mpqc_src_lib_math_scmat_predicate_h

#include <cmath>

namespace sc {

  /**
   * useful comparison functions
   */
  template <class T = double>
  struct abs_less : public std::binary_function<T, T, bool> {
    bool operator()(const T& a, const T& b) {
      return std::less<T>()(std::abs(a),std::abs(b));
    }
  };
  template <class T = double>
  struct abs_greater : public std::binary_function<T, T, bool> {
    bool operator()(const T& a, const T& b) {
      return std::greater<T>()(std::abs(a),std::abs(b));
    }
  };

} // end of namespace sc

#endif // end of header guard


// Local Variables:
// mode: c++
// c-file-style: "CLJ-CONDENSED"
// End:
