//
// Copyright (c) 2010
// Toon Knapen, Karl Meerbergen, Kresimir Fresl,
// Thomas Klimpel and Rutger ter Borg
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_NUMERIC_BINDINGS_LAPACK_DETAIL_CLAPACK_H
#define BOOST_NUMERIC_BINDINGS_LAPACK_DETAIL_CLAPACK_H

extern "C" {
#include <clapack.h> 
}

#endif 
